@@ -27,16 +27,7 @@ import './utils.js';
 import logger from './logger.js';
 import µb from './background.js';
 
-import {
-    StaticExtFilteringHostnameDB,
-    StaticExtFilteringSessionDB,
-} from './static-ext-filtering-db.js';
-
-/******************************************************************************/
-
-const cosmeticSurveyingMissCountMax =
-    parseInt(vAPI.localStorage.getItem('cosmeticSurveyingMissCountMax'), 10) ||
-    15;
+import { StaticExtFilteringHostnameDB } from './static-ext-filtering-db.js';
 
 /******************************************************************************/
 /******************************************************************************/
@@ -48,71 +39,55 @@ const SelectorCacheEntry = class {
 
     reset() {
         this.cosmetic = new Set();
-        this.cosmeticSurveyingMissCount = 0;
+        this.cosmeticHashes = new Set();
+        this.disableSurveyor = false;
         this.net = new Map();
-        this.lastAccessTime = Date.now();
+        this.accessId = SelectorCacheEntry.accessId++;
         return this;
     }
 
     dispose() {
-        this.cosmetic = this.net = null;
+        this.cosmetic = this.cosmeticHashes = this.net = null;
         if ( SelectorCacheEntry.junkyard.length < 25 ) {
             SelectorCacheEntry.junkyard.push(this);
         }
     }
 
     addCosmetic(details) {
-        const selectors = details.selectors;
-        let i = selectors.length || 0;
-        // https://github.com/gorhill/uBlock/issues/2011
-        //   Avoiding seemingly pointless surveys only if they appear costly.
-        if ( details.first && i === 0 ) {
-            if ( (details.cost || 0) >= 80 ) {
-                this.cosmeticSurveyingMissCount += 1;
-            }
-            return;
+        const selectors = details.selectors.join(',\n');
+        if ( selectors.length !== 0 ) {
+            this.cosmetic.add(selectors);
         }
-        this.cosmeticSurveyingMissCount = 0;
-        while ( i-- ) {
-            this.cosmetic.add(selectors[i]);
+        for ( const hash of details.hashes ) {
+            this.cosmeticHashes.add(hash);
         }
     }
 
     addNet(selectors) {
         if ( typeof selectors === 'string' ) {
-            this.addNetOne(selectors, Date.now());
+            this.net.set(selectors, this.accessId);
         } else {
-            this.addNetMany(selectors, Date.now());
+            this.net.set(selectors.join(',\n'), this.accessId);
         }
         // Net request-derived selectors: I limit the number of cached
-        // selectors, as I expect cases where the blocked net-requests
+        // selectors, as I expect cases where the blocked network requests
         // are never the exact same URL.
-        if ( this.net.size < SelectorCacheEntry.netHighWaterMark ) {
-            return;
-        }
-        const dict = this.net;
-        const keys = Array.from(dict.keys()).sort(function(a, b) {
-            return dict.get(b) - dict.get(a);
-        }).slice(SelectorCacheEntry.netLowWaterMark);
-        let i = keys.length;
-        while ( i-- ) {
-            dict.delete(keys[i]);
+        if ( this.net.size < SelectorCacheEntry.netHighWaterMark ) { return; }
+        const keys = Array.from(this.net)
+            .sort((a, b) => b[1] - a[1])
+            .slice(SelectorCacheEntry.netLowWaterMark)
+            .map(a => a[0]);
+        for ( const key of keys ) {
+            this.net.delete(key);
         }
     }
 
-    addNetOne(selector, now) {
-        this.net.set(selector, now);
-    }
-
-    addNetMany(selectors, now) {
-        let i = selectors.length || 0;
-        while ( i-- ) {
-            this.net.set(selectors[i], now);
-        }
+    addNetOne(selector, token) {
+        this.net.set(selector, token);
     }
 
     add(details) {
-        this.lastAccessTime = Date.now();
+        this.accessId = SelectorCacheEntry.accessId++;
         if ( details.type === 'cosmetic' ) {
             this.addCosmetic(details);
         } else {
@@ -122,10 +97,9 @@ const SelectorCacheEntry = class {
 
     // https://github.com/chrisaljoudi/uBlock/issues/420
     remove(type) {
-        this.lastAccessTime = Date.now();
+        this.accessId = SelectorCacheEntry.accessId++;
         if ( type === undefined || type === 'cosmetic' ) {
             this.cosmetic.clear();
-            this.cosmeticSurveyingMissCount = 0;
         }
         if ( type === undefined || type === 'net' ) {
             this.net.clear();
@@ -133,36 +107,41 @@ const SelectorCacheEntry = class {
     }
 
     retrieveToArray(iterator, out) {
-        for ( let selector of iterator ) {
+        for ( const selector of iterator ) {
             out.push(selector);
         }
     }
 
     retrieveToSet(iterator, out) {
-        for ( let selector of iterator ) {
+        for ( const selector of iterator ) {
             out.add(selector);
         }
     }
 
-    retrieve(type, out) {
-        this.lastAccessTime = Date.now();
-        const iterator = type === 'cosmetic' ? this.cosmetic : this.net.keys();
-        if ( Array.isArray(out) ) {
-            this.retrieveToArray(iterator, out);
-        } else {
-            this.retrieveToSet(iterator, out);
-        }
+    retrieveNet(out) {
+        this.accessId = SelectorCacheEntry.accessId++;
+        if ( this.net.size === 0 ) { return false; }
+        this.retrieveToArray(this.net.keys(), out);
+        return true;
+    }
+
+    retrieveCosmetic(selectors, hashes) {
+        this.accessId = SelectorCacheEntry.accessId++;
+        if ( this.cosmetic.size === 0 ) { return false; }
+        this.retrieveToSet(this.cosmetic, selectors);
+        this.retrieveToArray(this.cosmeticHashes, hashes);
+        return true;
     }
 
     static factory() {
         const entry = SelectorCacheEntry.junkyard.pop();
-        if ( entry ) {
-            return entry.reset();
-        }
-        return new SelectorCacheEntry();
+        return entry
+            ? entry.reset()
+            : new SelectorCacheEntry();
     }
 };
 
+SelectorCacheEntry.accessId = 1;
 SelectorCacheEntry.netLowWaterMark = 20;
 SelectorCacheEntry.netHighWaterMark = 30;
 SelectorCacheEntry.junkyard = [];
@@ -170,6 +149,63 @@ SelectorCacheEntry.junkyard = [];
 /******************************************************************************/
 /******************************************************************************/
 
+// http://www.cse.yorku.ca/~oz/hash.html#djb2
+//   Must mirror content script surveyor's version
+
+const hashFromStr = (type, s) => {
+    const len = s.length;
+    const step = len + 7 >>> 3;
+    let hash = (type << 5) + type ^ len;
+    for ( let i = 0; i < len; i += step ) {
+        hash = (hash << 5) + hash ^ s.charCodeAt(i);
+    }
+    return hash & 0xFFFFFF;
+};
+
+// https://github.com/gorhill/uBlock/issues/1668
+//   The key must be literal: unescape escaped CSS before extracting key.
+//   It's an uncommon case, so it's best to unescape only when needed.
+
+const keyFromSelector = selector => {
+    let key = '';
+    let matches = rePlainSelector.exec(selector);
+    if ( matches !== null ) {
+        key = matches[0];
+    } else {
+        matches = rePlainSelectorEx.exec(selector);
+        if ( matches === null ) { return; }
+        key = matches[1] || matches[2];
+    }
+    if ( key.includes('\\') === false ) { return key; }
+    matches = rePlainSelectorEscaped.exec(selector);
+    if ( matches === null ) { return; }
+    key = '';
+    const escaped = matches[0];
+    let beg = 0;
+    reEscapeSequence.lastIndex = 0;
+    for (;;) {
+        matches = reEscapeSequence.exec(escaped);
+        if ( matches === null ) {
+            return key + escaped.slice(beg);
+        }
+        key += escaped.slice(beg, matches.index);
+        beg = reEscapeSequence.lastIndex;
+        if ( matches[1].length === 1 ) {
+            key += matches[1];
+        } else {
+            key += String.fromCharCode(parseInt(matches[1], 16));
+        }
+    }
+};
+
+const rePlainSelector = /^[#.][\w\\-]+/;
+const rePlainSelectorEx = /^[^#.\[(]+([#.][\w-]+)|([#.][\w-]+)$/;
+const rePlainSelectorEscaped = /^[#.](?:\\[0-9A-Fa-f]+ |\\.|\w|-)+/;
+const reEscapeSequence = /\\([0-9A-Fa-f]+ |.)/g;
+
+/******************************************************************************/
+/******************************************************************************/
+
 // Cosmetic filter family tree:
 //
 // Generic
@@ -186,40 +222,19 @@ SelectorCacheEntry.junkyard = [];
 // Specific filers can be enforced before the main document is loaded.
 
 const FilterContainer = function() {
-    this.rePlainSelector = /^[#.][\w\\-]+/;
-    this.rePlainSelectorEscaped = /^[#.](?:\\[0-9A-Fa-f]+ |\\.|\w|-)+/;
-    this.rePlainSelectorEx = /^[^#.\[(]+([#.][\w-]+)|([#.][\w-]+)$/;
-    this.reEscapeSequence = /\\([0-9A-Fa-f]+ |.)/g;
     this.reSimpleHighGeneric = /^(?:[a-z]*\[[^\]]+\]|\S+)$/;
-    this.reHighMedium = /^\[href\^="https?:\/\/([^"]{8})[^"]*"\]$/;
 
     this.selectorCache = new Map();
     this.selectorCachePruneDelay = 10 * 60 * 1000; // 10 minutes
-    this.selectorCacheAgeMax = 120 * 60 * 1000; // 120 minutes
-    this.selectorCacheCountMin = 25;
-    this.netSelectorCacheCountMax = SelectorCacheEntry.netHighWaterMark;
+    this.selectorCacheCountMin = 40;
+    this.selectorCacheCountMax = 50;
     this.selectorCacheTimer = null;
 
     // specific filters
     this.specificFilters = new StaticExtFilteringHostnameDB(2);
 
-    // temporary filters
-    this.sessionFilterDB = new StaticExtFilteringSessionDB();
-
-    // low generic cosmetic filters, organized by id/class then simple/complex.
-    this.lowlyGeneric = Object.create(null);
-    this.lowlyGeneric.id = {
-        canonical: 'ids',
-        prefix: '#',
-        simple: new Set(),
-        complex: new Map()
-    };
-    this.lowlyGeneric.cl = {
-        canonical: 'classes',
-        prefix: '.',
-        simple: new Set(),
-        complex: new Map()
-    };
+    // low generic cosmetic filters: map of hash => stringified selector list
+    this.lowlyGeneric = new Map();
 
     // highly generic selectors sets
     this.highlyGeneric = Object.create(null);
@@ -240,8 +255,6 @@ const FilterContainer = function() {
     // is to prevent repeated allocation/deallocation overheads -- the
     // constructors/destructors of javascript Set/Map is assumed to be costlier
     // than just calling clear() on these.
-    this.$simpleSet = new Set();
-    this.$complexSet = new Set();
     this.$specificSet = new Set();
     this.$exceptionSet = new Set();
     this.$proceduralSet = new Set();
@@ -266,17 +279,11 @@ FilterContainer.prototype.reset = function() {
         this.selectorCacheTimer = null;
     }
 
-    // whether there is at least one surveyor-based filter
-    this.needDOMSurveyor = false;
-
     // hostname, entity-based filters
     this.specificFilters.clear();
 
-    // low generic cosmetic filters, organized by id/class then simple/complex.
-    this.lowlyGeneric.id.simple.clear();
-    this.lowlyGeneric.id.complex.clear();
-    this.lowlyGeneric.cl.simple.clear();
-    this.lowlyGeneric.cl.complex.clear();
+    // low generic cosmetic filters
+    this.lowlyGeneric.clear();
 
     // highly generic selectors sets
     this.highlyGeneric.simple.dict.clear();
@@ -285,6 +292,8 @@ FilterContainer.prototype.reset = function() {
     this.highlyGeneric.complex.dict.clear();
     this.highlyGeneric.complex.str = '';
     this.highlyGeneric.complex.mru.reset();
+
+    this.selfieVersion = 1;
 };
 
 /******************************************************************************/
@@ -293,12 +302,6 @@ FilterContainer.prototype.freeze = function() {
     this.duplicateBuster.clear();
     this.specificFilters.collectGarbage();
 
-    this.needDOMSurveyor =
-        this.lowlyGeneric.id.simple.size !== 0 ||
-        this.lowlyGeneric.id.complex.size !== 0 ||
-        this.lowlyGeneric.cl.simple.size !== 0 ||
-        this.lowlyGeneric.cl.complex.size !== 0;
-
     this.highlyGeneric.simple.str = Array.from(this.highlyGeneric.simple.dict).join(',\n');
     this.highlyGeneric.simple.mru.reset();
     this.highlyGeneric.complex.str = Array.from(this.highlyGeneric.complex.dict).join(',\n');
@@ -309,40 +312,6 @@ FilterContainer.prototype.freeze = function() {
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/1668
-//   The key must be literal: unescape escaped CSS before extracting key.
-//   It's an uncommon case, so it's best to unescape only when needed.
-
-FilterContainer.prototype.keyFromSelector = function(selector) {
-    let matches = this.rePlainSelector.exec(selector);
-    if ( matches === null ) { return; }
-    let key = matches[0];
-    if ( key.indexOf('\\') === -1 ) {
-        return key;
-    }
-    matches = this.rePlainSelectorEscaped.exec(selector);
-    if ( matches === null ) { return; }
-    key = '';
-    const escaped = matches[0];
-    let beg = 0;
-    this.reEscapeSequence.lastIndex = 0;
-    for (;;) {
-        matches = this.reEscapeSequence.exec(escaped);
-        if ( matches === null ) {
-            return key + escaped.slice(beg);
-        }
-        key += escaped.slice(beg, matches.index);
-        beg = this.reEscapeSequence.lastIndex;
-        if ( matches[1].length === 1 ) {
-            key += matches[1];
-        } else {
-            key += String.fromCharCode(parseInt(matches[1], 16));
-        }
-    }
-};
-
-/******************************************************************************/
-
 FilterContainer.prototype.compile = function(parser, writer) {
     if ( parser.hasOptions() === false ) {
         this.compileGenericSelector(parser, writer);
@@ -396,38 +365,8 @@ FilterContainer.prototype.compileGenericHideSelector = function(
 
     writer.select('COSMETIC_FILTERS:GENERIC');
 
-    const type = compiled.charCodeAt(0);
-    let key;
-
-    // Simple selector-based CSS rule: no need to test for whether the
-    // selector is valid, the regex took care of this. Most generic selector
-    // falls into that category:
-    // - ###ad-bigbox
-    // - ##.ads-bigbox
-    if ( type === 0x23 /* '#' */ ) {
-        key = this.keyFromSelector(compiled);
-        if ( key === compiled ) {
-            writer.push([ 0, key.slice(1) ]);
-            return;
-        }
-    } else if ( type === 0x2E /* '.' */ ) {
-        key = this.keyFromSelector(compiled);
-        if ( key === compiled ) {
-            writer.push([ 2, key.slice(1) ]);
-            return;
-        }
-    }
-
-    // Invalid cosmetic filter, possible reasons:
-    // - Bad syntax
-    // - Procedural filters (can't be generic): the compiled version of
-    //   a procedural selector is NEVER equal to its raw version.
-    // https://github.com/uBlockOrigin/uBlock-issues/issues/464
-    //   Pseudoclass-based selectors can be compiled, but are also valid
-    //   plain selectors.
     // https://github.com/uBlockOrigin/uBlock-issues/issues/131
     //   Support generic procedural filters as per advanced settings.
-    //   TODO: prevent double compilation.
     if ( compiled.charCodeAt(0) === 0x7B /* '{' */ ) {
         if ( µb.hiddenSettings.allowGenericProceduralFilters === true ) {
             return this.compileSpecificSelector(parser, '', false, writer);
@@ -441,28 +380,12 @@ FilterContainer.prototype.compileGenericHideSelector = function(
         return;
     }
 
-    // Complex selector-based CSS rule:
-    // - ###tads + div + .c
-    // - ##.rscontainer > .ellip
+    const key = keyFromSelector(compiled);
     if ( key !== undefined ) {
         writer.push([
-            type === 0x23 /* '#' */ ? 1 : 3,
-            key.slice(1),
-            compiled
-        ]);
-        return;
-    }
-
-    // https://github.com/gorhill/uBlock/issues/909
-    //   Anything which contains a plain id/class selector can be classified
-    //   as a low generic cosmetic filter.
-    const matches = this.rePlainSelectorEx.exec(compiled);
-    if ( matches !== null ) {
-        const key = matches[1] || matches[2];
-        writer.push([
-            key.charCodeAt(0) === 0x23 /* '#' */ ? 1 : 3,
-            key.slice(1),
-            compiled
+            0,
+            hashFromStr(key.charCodeAt(0), key.slice(1)),
+            compiled,
         ]);
         return;
     }
@@ -549,15 +472,6 @@ FilterContainer.prototype.compileSpecificSelector = function(
 
 /******************************************************************************/
 
-FilterContainer.prototype.compileTemporary = function(parser) {
-    return {
-        session: this.sessionFilterDB,
-        selector: parser.result.compiled,
-    };
-};
-
-/******************************************************************************/
-
 FilterContainer.prototype.fromCompiledContent = function(reader, options) {
     if ( options.skipCosmetic ) {
         this.skipCompiledContent(reader, 'SPECIFIC');
@@ -618,36 +532,13 @@ FilterContainer.prototype.fromCompiledContent = function(reader, options) {
         this.duplicateBuster.add(fingerprint);
         const args = reader.args();
         switch ( args[0] ) {
-        // low generic, simple
-        case 0:   // #AdBanner
-        case 2: { // .largeAd
-            const db = args[0] === 0 ? this.lowlyGeneric.id : this.lowlyGeneric.cl;
-            const bucket = db.complex.get(args[1]);
-            if ( bucket === undefined ) {
-                db.simple.add(args[1]);
-            } else if ( Array.isArray(bucket) ) {
-                bucket.push(db.prefix + args[1]);
+        // low generic
+        case 0: {
+            if ( this.lowlyGeneric.has(args[1]) ) {
+                const selector = this.lowlyGeneric.get(args[1]);
+                this.lowlyGeneric.set(args[1], `${selector},\n${args[2]}`);
             } else {
-                db.complex.set(args[1], [ bucket, db.prefix + args[1] ]);
-            }
-            break;
-        }
-        // low generic, complex
-        case 1:   // #tads + div + .c
-        case 3: { // .Mpopup + #Mad > #MadZone
-            const db = args[0] === 1 ? this.lowlyGeneric.id : this.lowlyGeneric.cl;
-            const bucket = db.complex.get(args[1]);
-            if ( bucket === undefined ) {
-                if ( db.simple.has(args[1]) ) {
-                    db.complex.set(args[1], [ db.prefix + args[1], args[2] ]);
-                } else {
-                    db.complex.set(args[1], args[2]);
-                    db.simple.add(args[1]);
-                }
-            } else if ( Array.isArray(bucket) ) {
-                bucket.push(args[2]);
-            } else {
-                db.complex.set(args[1], [ bucket, args[2] ]);
+                this.lowlyGeneric.set(args[1], args[2]);
             }
             break;
         }
@@ -682,13 +573,11 @@ FilterContainer.prototype.skipCompiledContent = function(reader, sectionId) {
 
 FilterContainer.prototype.toSelfie = function() {
     return {
+        version: this.selfieVersion,
         acceptedCount: this.acceptedCount,
         discardedCount: this.discardedCount,
         specificFilters: this.specificFilters.toSelfie(),
-        lowlyGenericSID: Array.from(this.lowlyGeneric.id.simple),
-        lowlyGenericCID: Array.from(this.lowlyGeneric.id.complex),
-        lowlyGenericSCL: Array.from(this.lowlyGeneric.cl.simple),
-        lowlyGenericCCL: Array.from(this.lowlyGeneric.cl.complex),
+        lowlyGeneric: Array.from(this.lowlyGeneric),
         highSimpleGenericHideArray: Array.from(this.highlyGeneric.simple.dict),
         highComplexGenericHideArray: Array.from(this.highlyGeneric.complex.dict),
     };
@@ -697,22 +586,19 @@ FilterContainer.prototype.toSelfie = function() {
 /******************************************************************************/
 
 FilterContainer.prototype.fromSelfie = function(selfie) {
+    if ( selfie.version !== this.selfieVersion ) {
+        throw new Error(
+            `cosmeticFilteringEngine: mismatched selfie version, ${selfie.version}, expected ${this.selfieVersion}`
+        );
+    }
     this.acceptedCount = selfie.acceptedCount;
     this.discardedCount = selfie.discardedCount;
     this.specificFilters.fromSelfie(selfie.specificFilters);
-    this.lowlyGeneric.id.simple = new Set(selfie.lowlyGenericSID);
-    this.lowlyGeneric.id.complex = new Map(selfie.lowlyGenericCID);
-    this.lowlyGeneric.cl.simple = new Set(selfie.lowlyGenericSCL);
-    this.lowlyGeneric.cl.complex = new Map(selfie.lowlyGenericCCL);
+    this.lowlyGeneric = new Map(selfie.lowlyGeneric);
     this.highlyGeneric.simple.dict = new Set(selfie.highSimpleGenericHideArray);
     this.highlyGeneric.simple.str = selfie.highSimpleGenericHideArray.join(',\n');
     this.highlyGeneric.complex.dict = new Set(selfie.highComplexGenericHideArray);
     this.highlyGeneric.complex.str = selfie.highComplexGenericHideArray.join(',\n');
-    this.needDOMSurveyor =
-        selfie.lowlyGenericSID.length !== 0 ||
-        selfie.lowlyGenericCID.length !== 0 ||
-        selfie.lowlyGenericSCL.length !== 0 ||
-        selfie.lowlyGenericCCL.length !== 0;
     this.frozen = true;
 };
 
@@ -721,12 +607,11 @@ FilterContainer.prototype.fromSelfie = function(selfie) {
 FilterContainer.prototype.triggerSelectorCachePruner = function() {
     // Of interest: http://fitzgeraldnick.com/weblog/40/
     // http://googlecode.blogspot.ca/2009/07/gmail-for-mobile-html5-series-using.html
-    if ( this.selectorCacheTimer === null ) {
-        this.selectorCacheTimer = vAPI.setTimeout(
-            this.pruneSelectorCacheAsync.bind(this),
-            this.selectorCachePruneDelay
-        );
-    }
+    if ( this.selectorCacheTimer !== null ) { return; }
+    this.selectorCacheTimer = vAPI.setTimeout(
+        ( ) => { this.pruneSelectorCacheAsync(); },
+        this.selectorCachePruneDelay
+    );
 };
 
 /******************************************************************************/
@@ -740,7 +625,7 @@ FilterContainer.prototype.addToSelectorCache = function(details) {
     if ( entry === undefined ) {
         entry = SelectorCacheEntry.factory();
         this.selectorCache.set(hostname, entry);
-        if ( this.selectorCache.size > this.selectorCacheCountMin ) {
+        if ( this.selectorCache.size > this.selectorCacheCountMax ) {
             this.triggerSelectorCachePruner();
         }
     }
@@ -753,7 +638,7 @@ FilterContainer.prototype.removeFromSelectorCache = function(
     targetHostname = '*',
     type = undefined
 ) {
-    let targetHostnameLength = targetHostname.length;
+    const targetHostnameLength = targetHostname.length;
     for ( let entry of this.selectorCache ) {
         let hostname = entry[0];
         let item = entry[1];
@@ -772,52 +657,27 @@ FilterContainer.prototype.removeFromSelectorCache = function(
 
 /******************************************************************************/
 
-FilterContainer.prototype.retrieveFromSelectorCache = function(
-    hostname,
-    type,
-    out
-) {
-    let entry = this.selectorCache.get(hostname);
-    if ( entry !== undefined ) {
-        entry.retrieve(type, out);
-    }
-};
-
-/******************************************************************************/
-
 FilterContainer.prototype.pruneSelectorCacheAsync = function() {
     this.selectorCacheTimer = null;
-    if ( this.selectorCache.size <= this.selectorCacheCountMin ) { return; }
-    let cache = this.selectorCache;
-    // Sorted from most-recently-used to least-recently-used, because
-    //   we loop beginning at the end below.
-    // We can't avoid sorting because we have to keep a minimum number of
-    //   entries, and these entries should always be the most-recently-used.
-    let hostnames = Array.from(cache.keys())
-            .sort(function(a, b) {
-                return cache.get(b).lastAccessTime -
-                       cache.get(a).lastAccessTime;
-                })
-            .slice(this.selectorCacheCountMin);
-    let obsolete = Date.now() - this.selectorCacheAgeMax,
-        i = hostnames.length;
-    while ( i-- ) {
-        let hostname = hostnames[i];
-        let entry = cache.get(hostname);
-        if ( entry.lastAccessTime > obsolete ) { break; }
-        // console.debug('pruneSelectorCacheAsync: flushing "%s"', hostname);
-        entry.dispose();
-        cache.delete(hostname);
-    }
-    if ( cache.size > this.selectorCacheCountMin ) {
-        this.triggerSelectorCachePruner();
+    if ( this.selectorCache.size <= this.selectorCacheCountMax ) { return; }
+    const cache = this.selectorCache;
+    const hostnames = Array.from(cache.keys())
+        .sort((a, b) => cache.get(b).accessId - cache.get(a).accessId)
+        .slice(this.selectorCacheCountMin);
+    for ( const hn of hostnames ) {
+        cache.get(hn).dispose();
+        cache.delete(hn);
     }
 };
 
 /******************************************************************************/
 
-FilterContainer.prototype.getSession = function() {
-    return this.sessionFilterDB;
+FilterContainer.prototype.disableSurveyor = function(details) {
+    const hostname = details.hostname;
+    if ( typeof hostname !== 'string' || hostname === '' ) { return; }
+    const cacheEntry = this.selectorCache.get(hostname);
+    if ( cacheEntry === undefined ) { return; }
+    cacheEntry.disableSurveyor = true;
 };
 
 /******************************************************************************/
@@ -850,92 +710,51 @@ FilterContainer.prototype.cssRuleFromProcedural = function(json) {
 /******************************************************************************/
 
 FilterContainer.prototype.retrieveGenericSelectors = function(request) {
-    if ( this.acceptedCount === 0 ) { return; }
-    if ( !request.ids && !request.classes ) { return; }
+    if ( this.lowlyGeneric.size === 0 ) { return; }
+    if ( Array.isArray(request.hashes) === false ) { return; }
+    if ( request.hashes.length === 0 ) { return; }
 
-    const { safeOnly = false } = request;
-    //console.time('cosmeticFilteringEngine.retrieveGenericSelectors');
-
-    const simpleSelectors = this.$simpleSet;
-    const complexSelectors = this.$complexSet;
-
-    const cacheEntry = this.selectorCache.get(request.hostname);
-    const previousHits = cacheEntry && cacheEntry.cosmetic || this.$dummySet;
-
-    for ( const type in this.lowlyGeneric ) {
-        const entry = this.lowlyGeneric[type];
-        const selectors = request[entry.canonical];
-        if ( Array.isArray(selectors) === false ) { continue; }
-        for ( const identifier of selectors ) {
-            if ( entry.simple.has(identifier) === false ) { continue; }
-            const bucket = entry.complex.get(identifier);
-            if ( typeof bucket === 'string' ) {
-                if ( previousHits.has(bucket) ) { continue; }
-                complexSelectors.add(bucket);
-                continue;
-            }
-            const simpleSelector = entry.prefix + identifier;
-            if ( Array.isArray(bucket) ) {
-                for ( const complexSelector of bucket ) {
-                    if ( previousHits.has(complexSelector) ) { continue; }
-                    if ( safeOnly && complexSelector === simpleSelector ) { continue; }
-                    complexSelectors.add(complexSelector);
-                }
-                continue;
-            }
-            if ( previousHits.has(simpleSelector) ) { continue; }
-            if ( safeOnly ) { continue; }
-            simpleSelectors.add(simpleSelector);
+    const selectorsSet = new Set();
+    const hashes = [];
+    const safeOnly = request.safeOnly === true;
+    for ( const hash of request.hashes ) {
+        const bucket = this.lowlyGeneric.get(hash);
+        if ( bucket === undefined ) { continue; }
+        for ( const selector of bucket.split(',\n') ) {
+            if ( safeOnly && selector === keyFromSelector(selector) ) { continue; }
+            selectorsSet.add(selector);
         }
+        hashes.push(hash);
     }
 
     // Apply exceptions: it is the responsibility of the caller to provide
     // the exceptions to be applied.
     const excepted = [];
-    if ( Array.isArray(request.exceptions) ) {
+    if ( selectorsSet.size !== 0 && Array.isArray(request.exceptions) ) {
         for ( const exception of request.exceptions ) {
-            if (
-                simpleSelectors.delete(exception) ||
-                complexSelectors.delete(exception)
-            ) {
+            if ( selectorsSet.delete(exception) ) {
                 excepted.push(exception);
             }
         }
     }
 
-    if (
-        simpleSelectors.size === 0 &&
-        complexSelectors.size === 0 &&
-        excepted.length === 0
-    ) {
-        return;
-    }
+    if ( selectorsSet.size === 0 && excepted.length === 0 ) { return; }
 
     const out = { injectedCSS: '', excepted, };
-
-    const injected = [];
-    if ( simpleSelectors.size !== 0 ) {
-        injected.push(...simpleSelectors);
-        simpleSelectors.clear();
-    }
-    if ( complexSelectors.size !== 0 ) {
-        injected.push(...complexSelectors);
-        complexSelectors.clear();
-    }
-
-    // Cache and inject looked-up low generic cosmetic filters.
-    if ( injected.length === 0 ) { return out; }
+    const selectors = Array.from(selectorsSet);
 
     if ( typeof request.hostname === 'string' && request.hostname !== '' ) {
         this.addToSelectorCache({
-            cost: request.surveyCost || 0,
             hostname: request.hostname,
-            selectors: injected,
+            selectors,
+            hashes,
             type: 'cosmetic',
         });
     }
 
-    out.injectedCSS = `${injected.join(',\n')}\n{display:none!important;}`;
+    if ( selectors.length === 0 ) { return out; }
+
+    out.injectedCSS = `${selectors.join(',\n')}\n{display:none!important;}`;
     vAPI.tabs.insertCSS(request.tabId, {
         code: out.injectedCSS,
         frameId: request.frameId,
@@ -943,8 +762,6 @@ FilterContainer.prototype.retrieveGenericSelectors = function(request) {
         runAt: 'document_start',
     });
 
-    //console.timeEnd('cosmeticFilteringEngine.retrieveGenericSelectors');
-
     return out;
 };
 
@@ -972,7 +789,7 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         exceptedFilters: [],
         proceduralFilters: [],
         convertedProceduralFilters: [],
-        noDOMSurveying: this.needDOMSurveyor === false,
+        disableSurveyor: this.lowlyGeneric.size === 0,
     };
     const injectedCSS = [];
 
@@ -980,7 +797,6 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         options.noSpecificCosmeticFiltering !== true ||
         options.noGenericCosmeticFiltering !== true
     ) {
-        const injectedHideFilters = [];
         const specificSet = this.$specificSet;
         const proceduralSet = this.$proceduralSet;
         const exceptionSet = this.$exceptionSet;
@@ -988,18 +804,12 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
 
         // Cached cosmetic filters: these are always declarative.
         if ( cacheEntry !== undefined ) {
-            cacheEntry.retrieve('cosmetic', specificSet);
-            if ( out.noDOMSurveying === false ) {
-                out.noDOMSurveying = cacheEntry.cosmeticSurveyingMissCount >
-                                   cosmeticSurveyingMissCountMax;
+            cacheEntry.retrieveCosmetic(specificSet, out.genericCosmeticHashes = []);
+            if ( cacheEntry.disableSurveyor ) {
+                out.disableSurveyor = true;
             }
         }
 
-        // Retrieve temporary filters
-        if ( this.sessionFilterDB.isNotEmpty ) {
-            this.sessionFilterDB.retrieve([ null, exceptionSet ]);
-        }
-
         // Retrieve filters with a non-empty hostname
         this.specificFilters.retrieve(
             hostname,
@@ -1040,7 +850,9 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         }
 
         if ( specificSet.size !== 0 ) {
-            injectedHideFilters.push(Array.from(specificSet).join(',\n'));
+            injectedCSS.push(
+                `${Array.from(specificSet).join(',\n')}\n{display:none!important;}`
+            );
         }
 
         // Some procedural filters are really declarative cosmetic filters, so
@@ -1091,17 +903,11 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
                     out.exceptedFilters.push(...str.excepted);
                 }
                 if ( str.s.length !== 0 ) {
-                    injectedHideFilters.push(str.s);
+                    injectedCSS.push(`${str.s}\n{display:none!important;}`);
                 }
             }
         }
 
-        if ( injectedHideFilters.length !== 0 ) {
-            injectedCSS.push(
-                `${injectedHideFilters.join(',\n')}\n{display:none!important;}`
-            );
-        }
-
         // Important: always clear used registers before leaving.
         specificSet.clear();
         proceduralSet.clear();
@@ -1128,9 +934,8 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
     // CSS selectors for collapsible blocked elements
     if ( cacheEntry ) {
         const networkFilters = [];
-        cacheEntry.retrieve('net', networkFilters);
-        if ( networkFilters.length !== 0 ) {
-            details.code = networkFilters.join('\n') + '\n{display:none!important;}';
+        if ( cacheEntry.retrieveNet(networkFilters) ) {
+            details.code = `${networkFilters.join('\n')}\n{display:none!important;}`;
             if ( request.tabId !== undefined ) {
                 vAPI.tabs.insertCSS(request.tabId, details);
             }
@@ -1149,31 +954,16 @@ FilterContainer.prototype.getFilterCount = function() {
 /******************************************************************************/
 
 FilterContainer.prototype.dump = function() {
-    let genericCount = 0;
-    for ( const i of [ 'simple', 'complex' ] ) {
-        for ( const j of [ 'id', 'cl' ] ) {
-            genericCount += this.lowlyGeneric[j][i].size;
-        }
+    const generics = [];
+    for ( const selectors of this.lowlyGeneric.values() ) {
+        generics.push(...selectors.split(',\n'));
     }
     return [
         'Cosmetic Filtering Engine internals:',
         `specific: ${this.specificFilters.size}`,
-        `generic: ${genericCount}`,
-        `+ lowly.id: ${this.lowlyGeneric.id.simple.size + this.lowlyGeneric.id.complex.size}`,
-        `  + simple: ${this.lowlyGeneric.id.simple.size}`,
-        ...Array.from(this.lowlyGeneric.id.simple).map(a => `    ###${a}`),
-        `  + complex: ${this.lowlyGeneric.id.complex.size}`,
-        ...Array.from(this.lowlyGeneric.id.complex.values()).map(a => `    ##${a}`),
-        `+ lowly.class: ${this.lowlyGeneric.cl.simple.size + this.lowlyGeneric.cl.complex.size}`,
-        `  + simple: ${this.lowlyGeneric.cl.simple.size}`,
-        ...Array.from(this.lowlyGeneric.cl.simple).map(a => `    ##.${a}`),
-        `  + complex: ${this.lowlyGeneric.cl.complex.size}`,
-        ...Array.from(this.lowlyGeneric.cl.complex.values()).map(a => `    ##${a}`),
-        `+ highly: ${this.highlyGeneric.simple.dict.size + this.highlyGeneric.complex.dict.size}`,
-        `  + highly.simple: ${this.highlyGeneric.simple.dict.size}`,
-        ...Array.from(this.highlyGeneric.simple.dict).map(a => `    ##${a}`),
-        `  + highly.complex: ${this.highlyGeneric.complex.dict.size}`,
-        ...Array.from(this.highlyGeneric.complex.dict).map(a => `    ##${a}`),
+        `generic: ${generics.length}`,
+        `+ selectors: ${this.lowlyGeneric.size}`,
+        ...generics.map(a => `  ${a}`),
     ].join('\n');
 };
 
